@@ -13,9 +13,26 @@
 #ifndef GDAL_RAT_H_INCLUDED
 #define GDAL_RAT_H_INCLUDED
 
+#if !defined(GDAL_COMPILATION) &&                                              \
+    !defined(GDAL_RAT_SKIP_OTHER_GDAL_HEADERS) && !defined(GDAL_4_0_COMPAT)
+
 #include "cpl_minixml.h"
 #include "gdal_priv.h"
 
+#else
+
+#include "cpl_port.h"
+#include "cpl_error.h"
+#include "cpl_string.h"
+#include "gdal.h"
+
+#endif
+
+#include <memory>
+#include <vector>
+
+class GDALColorTable;
+
 // Clone and Serialize are allowed to fail if GetRowCount()*GetColCount()
 // greater than this number
 #define RAT_MAX_ELEM_FOR_CLONE 1000000
@@ -114,7 +131,9 @@ class CPL_DLL GDALRasterAttributeTable
      * \brief Fetch field value as a string.
      *
      * The value of the requested column in the requested row is returned
-     * as a string.  If the field is numeric, it is formatted as a string
+     * as a string. This method is nominally called on fields of type GFT_String,
+     * but it can be called on fields of other types as well.
+     * If the field is numeric, it is formatted as a string
      * using default rules, so some precision may be lost.
      *
      * The returned string is temporary and cannot be expected to be
@@ -133,8 +152,10 @@ class CPL_DLL GDALRasterAttributeTable
      * \brief Fetch field value as a integer.
      *
      * The value of the requested column in the requested row is returned
-     * as an integer.  Non-integer fields will be converted to integer with
-     * the possibility of data loss.
+     * as an integer. This method is nominally called on fields of type
+     * GFT_Integer, but it can be called on fields of other types as well.
+     * Non-integer fields will be converted to integer with the possibility of
+     * data loss.
      *
      * This method is the same as the C function GDALRATGetValueAsInt().
      *
@@ -149,7 +170,9 @@ class CPL_DLL GDALRasterAttributeTable
      * \brief Fetch field value as a double.
      *
      * The value of the requested column in the requested row is returned
-     * as a double.   Non double fields will be converted to double with
+     * as a double. This method is nominally called on fields of type
+     * GFT_Real, but it can be called on fields of other types as well.
+     * Non double fields will be converted to double with
      * the possibility of data loss.
      *
      * This method is the same as the C function GDALRATGetValueAsDouble().
@@ -161,11 +184,72 @@ class CPL_DLL GDALRasterAttributeTable
      */
     virtual double GetValueAsDouble(int iRow, int iField) const = 0;
 
+    /**
+     * \brief Fetch field value as a boolean.
+     *
+     * The value of the requested column in the requested row is returned
+     * as a boolean. This method is nominally called on fields of type
+     * GFT_Boolean, but it can be called on fields of other types as well.
+     * Non boolean fields will be converted to boolean with the possibility of
+     * data loss.
+     *
+     * This method is the same as the C function GDALRATGetValueAsBoolean().
+     *
+     * @param iRow row to fetch (zero based).
+     * @param iField column to fetch (zero based).
+     *
+     * @return field value
+     * @since 3.12
+     */
+    virtual bool GetValueAsBoolean(int iRow, int iField) const = 0;
+
+    /**
+     * \brief Fetch field value as a datetime.
+     *
+     * The value of the requested column in the requested row is returned
+     * as a datetime. Besides being called on a GFT_DateTime field, it
+     * is also possible to call this method on a string field that contains a
+     * ISO-8601 encoded datetime.
+     *
+     * This method is the same as the C function GDALRATGetValueAsDateTime().
+     *
+     * @param iRow row to fetch (zero based).
+     * @param iField column to fetch (zero based).
+     *
+     * @return field value
+     * @since 3.12
+     */
+    virtual GDALRATDateTime GetValueAsDateTime(int iRow, int iField) const = 0;
+
+    /**
+     * \brief Fetch field value as a WKB geometry.
+     *
+     * The value of the requested column in the requested row is returned
+     * as a WKB geometry. Besides being called on a GFT_WKBGeometry field, it
+     * is also possible to call this method on a string field that contains a WKT
+     * encoded geometry.
+     *
+     * The returned pointer may be invalidated by a following call to a method
+     * of this GDALRasterAttributeTable instance.
+     *
+     * This method is the same as the C function GDALRATGetValueAsWKBGeometry().
+     *
+     * @param iRow row to fetch (zero based).
+     * @param iField column to fetch (zero based).
+     * @param[out] nWKBSize Number of bytes of the returned pointer
+     * @return field value, or nullptr
+     * @since 3.12
+     */
+    virtual const GByte *GetValueAsWKBGeometry(int iRow, int iField,
+                                               size_t &nWKBSize) const = 0;
+
     /**
      * \brief Set field value from string.
      *
      * The indicated field (column) on the indicated row is set from the
-     * passed value.  The value will be automatically converted for other field
+     * passed value. This method is nominally called on fields of type
+     * GFT_String, but it can be called on fields of other types as well.
+     * The value will be automatically converted for other field
      * types, with a possible loss of precision.
      *
      * This method is the same as the C function GDALRATSetValueAsString().
@@ -173,14 +257,17 @@ class CPL_DLL GDALRasterAttributeTable
      * @param iRow row to fetch (zero based).
      * @param iField column to fetch (zero based).
      * @param pszValue the value to assign.
+     * @return (since 3.12) CE_None in case of success, error code otherwise
      */
-    virtual void SetValue(int iRow, int iField, const char *pszValue) = 0;
+    virtual CPLErr SetValue(int iRow, int iField, const char *pszValue) = 0;
 
     /**
      * \brief Set field value from integer.
      *
      * The indicated field (column) on the indicated row is set from the
-     * passed value.  The value will be automatically converted for other field
+     * passed value. This method is nominally called on fields of type
+     * GFT_Integer, but it can be called on fields of other types as well.
+     * The value will be automatically converted for other field
      * types, with a possible loss of precision.
      *
      * This method is the same as the C function GDALRATSetValueAsInteger().
@@ -188,14 +275,17 @@ class CPL_DLL GDALRasterAttributeTable
      * @param iRow row to fetch (zero based).
      * @param iField column to fetch (zero based).
      * @param nValue the value to assign.
+     * @return (since 3.12) CE_None in case of success, error code otherwise
      */
-    virtual void SetValue(int iRow, int iField, int nValue) = 0;
+    virtual CPLErr SetValue(int iRow, int iField, int nValue) = 0;
 
     /**
      * \brief Set field value from double.
      *
      * The indicated field (column) on the indicated row is set from the
-     * passed value.  The value will be automatically converted for other field
+     * passed value. This method is nominally called on fields of type
+     * GFT_Real, but it can be called on fields of other types as well.
+     * The value will be automatically converted for other field
      * types, with a possible loss of precision.
      *
      * This method is the same as the C function GDALRATSetValueAsDouble().
@@ -203,8 +293,70 @@ class CPL_DLL GDALRasterAttributeTable
      * @param iRow row to fetch (zero based).
      * @param iField column to fetch (zero based).
      * @param dfValue the value to assign.
+     * @return (since 3.12) CE_None in case of success, error code otherwise
      */
-    virtual void SetValue(int iRow, int iField, double dfValue) = 0;
+    virtual CPLErr SetValue(int iRow, int iField, double dfValue) = 0;
+
+    /**
+     * \brief Set field value from boolean.
+     *
+     * The indicated field (column) on the indicated row is set from the
+     * passed value.  This method is nominally called on fields of type
+     * GFT_Boolean, but it can be called on fields of other types as well.
+     * The value will be automatically converted for other field
+     * types, with a possible loss of precision.
+     *
+     * This method is the same as the C function GDALRATSetValueAsBoolean().
+     *
+     * @param iRow row to fetch (zero based).
+     * @param iField column to fetch (zero based).
+     * @param bValue the value to assign.
+     * @return CE_None in case of success, error code otherwise
+     * @since 3.12
+     */
+    virtual CPLErr SetValue(int iRow, int iField, bool bValue) = 0;
+
+    /**
+     * \brief Set field value from datetime.
+     *
+     * The indicated field (column) on the indicated row is set from the
+     * passed value. Besides being called on a field of type GFT_DateTime, this
+     * method can also be called on a field of type GFT_String, in which case
+     * the datetime will be converted into its ISO-8601 representation.
+     *
+     * Note that the GDALRATDateTime::bIsValid field must be set to true if
+     * the date time is valid.
+     *
+     * This method is the same as the C function GDALRATSetValueAsDateTime().
+     *
+     * @param iRow row to fetch (zero based).
+     * @param iField column to fetch (zero based).
+     * @param sDateTime Date time value
+     * @return CE_None in case of success, error code otherwise
+     * @since 3.12
+     */
+    virtual CPLErr SetValue(int iRow, int iField,
+                            const GDALRATDateTime &sDateTime) = 0;
+
+    /**
+     * \brief Set field value from a WKB geometry.
+     *
+     * The indicated field (column) on the indicated row is set from the
+     * passed value. Besides being called on a field of type GFT_WKBGeometry, this
+     * method can also be called on a field of type GFT_String, in which case
+     * the datetime will be converted into its WKT geometry representation.
+     *
+     * This method is the same as the C function GDALRATSetValueAsWKBGeometry().
+     *
+     * @param iRow row to fetch (zero based).
+     * @param iField column to fetch (zero based).
+     * @param pabyWKB Pointer to a WKB encoded geometry
+     * @param nWKBSize Number of bytes of pabyWKB.
+     * @return CE_None in case of success, error code otherwise
+     * @since 3.12
+     */
+    virtual CPLErr SetValue(int iRow, int iField, const void *pabyWKB,
+                            size_t nWKBSize) = 0;
 
     /**
      * \brief Determine whether changes made to this RAT are reflected directly
@@ -225,7 +377,6 @@ class CPL_DLL GDALRasterAttributeTable
      *
      * Set whether the RAT is thematic or athematic (continuous).
      *
-     * @since GDAL 2.4
      */
     virtual CPLErr SetTableType(const GDALRATTableType eInTableType) = 0;
 
@@ -234,7 +385,6 @@ class CPL_DLL GDALRasterAttributeTable
      *
      * Indicates whether the RAT is thematic or athematic (continuous).
      *
-     * @since GDAL 2.4
      * @return table type
      */
     virtual GDALRATTableType GetTableType() const = 0;
@@ -245,6 +395,12 @@ class CPL_DLL GDALRasterAttributeTable
                             int iLength, int *pnData);
     virtual CPLErr ValuesIO(GDALRWFlag eRWFlag, int iField, int iStartRow,
                             int iLength, char **papszStrList);
+    virtual CPLErr ValuesIO(GDALRWFlag eRWFlag, int iField, int iStartRow,
+                            int iLength, bool *pbData);
+    virtual CPLErr ValuesIO(GDALRWFlag eRWFlag, int iField, int iStartRow,
+                            int iLength, GDALRATDateTime *pasDateTime);
+    virtual CPLErr ValuesIO(GDALRWFlag eRWFlag, int iField, int iStartRow,
+                            int iLength, GByte **ppabyWKB, size_t *pnWKBSize);
 
     virtual void SetRowCount(int iCount);
     virtual int GetRowOfValue(double dfValue) const;
@@ -272,7 +428,6 @@ class CPL_DLL GDALRasterAttributeTable
     virtual void DumpReadable(FILE * = nullptr);
 
     /** Convert a GDALRasterAttributeTable* to a GDALRasterAttributeTableH.
-     * @since GDAL 2.3
      */
     static inline GDALRasterAttributeTableH
     ToHandle(GDALRasterAttributeTable *poRAT)
@@ -281,7 +436,6 @@ class CPL_DLL GDALRasterAttributeTable
     }
 
     /** Convert a GDALRasterAttributeTableH to a GDALRasterAttributeTable*.
-     * @since GDAL 2.3
      */
     static inline GDALRasterAttributeTable *
     FromHandle(GDALRasterAttributeTableH hRAT)
@@ -292,33 +446,39 @@ class CPL_DLL GDALRasterAttributeTable
     /**
      * \brief Remove statistics from the RAT.
      *
-     * @since GDAL 2.4
      */
     virtual void RemoveStatistics() = 0;
+
+    //! @cond Doxygen_Suppress
+    static std::string DateTimeToString(const GDALRATDateTime &sDateTime);
+    static bool StringToDateTime(const char *pszStr,
+                                 GDALRATDateTime &sDateTime);
+
+    static std::string WKBGeometryToWKT(const void *pabyWKB, size_t nWKBSize);
+    static std::vector<GByte> WKTGeometryToWKB(const char *pszWKT);
+    //! @endcond
+
+  protected:
+    //! @cond Doxygen_Suppress
+    GDALRasterAttributeTable() = default;
+    GDALRasterAttributeTable(const GDALRasterAttributeTable &) = default;
+    GDALRasterAttributeTable &
+    operator=(const GDALRasterAttributeTable &) = default;
+    GDALRasterAttributeTable(GDALRasterAttributeTable &&) = default;
+    GDALRasterAttributeTable &operator=(GDALRasterAttributeTable &&) = default;
+
+    CPLErr ValuesIOBooleanFromIntoInt(GDALRWFlag eRWFlag, int iField,
+                                      int iStartRow, int iLength, bool *pbData);
+    CPLErr ValuesIODateTimeFromIntoString(GDALRWFlag eRWFlag, int iField,
+                                          int iStartRow, int iLength,
+                                          GDALRATDateTime *psDateTime);
+    CPLErr ValuesIOWKBGeometryFromIntoString(GDALRWFlag eRWFlag, int iField,
+                                             int iStartRow, int iLength,
+                                             GByte **ppabyWKB,
+                                             size_t *pnWKBSize);
+    //! @endcond
 };
 
-/************************************************************************/
-/*                       GDALRasterAttributeField                       */
-/*                                                                      */
-/*      (private)                                                       */
-/************************************************************************/
-//! @cond Doxygen_Suppress
-class GDALRasterAttributeField
-{
-  public:
-    CPLString sName{};
-
-    GDALRATFieldType eType = GFT_Integer;
-
-    GDALRATFieldUsage eUsage = GFU_Generic;
-
-    std::vector<GInt32> anValues{};
-    std::vector<double> adfValues{};
-    std::vector<CPLString> aosValues{};
-};
-
-//! @endcond
-
 /************************************************************************/
 /*                    GDALDefaultRasterAttributeTable                   */
 /************************************************************************/
@@ -328,13 +488,29 @@ class GDALRasterAttributeField
 class CPL_DLL GDALDefaultRasterAttributeTable : public GDALRasterAttributeTable
 {
   private:
+    struct GDALRasterAttributeField
+    {
+        CPLString sName{};
+
+        GDALRATFieldType eType = GFT_Integer;
+
+        GDALRATFieldUsage eUsage = GFU_Generic;
+
+        std::vector<GInt32> anValues{};
+        std::vector<double> adfValues{};
+        std::vector<CPLString> aosValues{};
+        std::vector<bool> abValues{};
+        std::vector<GDALRATDateTime> asDateTimeValues{};
+        std::vector<std::vector<GByte>> aabyWKBGeometryValues{};
+    };
+
     std::vector<GDALRasterAttributeField> aoFields{};
 
     int bLinearBinning = false;  // TODO(schwehr): Can this be a bool?
     double dfRow0Min = -0.5;
     double dfBinSize = 1.0;
 
-    GDALRATTableType eTableType;
+    GDALRATTableType eTableType = GRTT_THEMATIC;
 
     void AnalyseColumns();
     int bColumnsAnalysed = false;  // TODO(schwehr): Can this be a bool?
@@ -344,11 +520,23 @@ class CPL_DLL GDALDefaultRasterAttributeTable : public GDALRasterAttributeTable
     int nRowCount = 0;
 
     CPLString osWorkingResult{};
+    mutable std::vector<GByte> m_abyWKB{};
 
   public:
     GDALDefaultRasterAttributeTable();
     ~GDALDefaultRasterAttributeTable() override;
 
+    //! @cond Doxygen_Suppress
+    GDALDefaultRasterAttributeTable(const GDALDefaultRasterAttributeTable &) =
+        default;
+    GDALDefaultRasterAttributeTable &
+    operator=(const GDALDefaultRasterAttributeTable &) = default;
+    GDALDefaultRasterAttributeTable(GDALDefaultRasterAttributeTable &&) =
+        default;
+    GDALDefaultRasterAttributeTable &
+    operator=(GDALDefaultRasterAttributeTable &&) = default;
+    //! @endcond
+
     GDALDefaultRasterAttributeTable *Clone() const override;
 
     int GetColumnCount() const override;
@@ -364,10 +552,19 @@ class CPL_DLL GDALDefaultRasterAttributeTable : public GDALRasterAttributeTable
     const char *GetValueAsString(int iRow, int iField) const override;
     int GetValueAsInt(int iRow, int iField) const override;
     double GetValueAsDouble(int iRow, int iField) const override;
+    bool GetValueAsBoolean(int iRow, int iField) const override;
+    GDALRATDateTime GetValueAsDateTime(int iRow, int iField) const override;
+    const GByte *GetValueAsWKBGeometry(int iRow, int iField,
+                                       size_t &nWKBSize) const override;
 
-    void SetValue(int iRow, int iField, const char *pszValue) override;
-    void SetValue(int iRow, int iField, double dfValue) override;
-    void SetValue(int iRow, int iField, int nValue) override;
+    CPLErr SetValue(int iRow, int iField, const char *pszValue) override;
+    CPLErr SetValue(int iRow, int iField, double dfValue) override;
+    CPLErr SetValue(int iRow, int iField, int nValue) override;
+    CPLErr SetValue(int iRow, int iField, bool bValue) override;
+    CPLErr SetValue(int iRow, int iField,
+                    const GDALRATDateTime &sDateTime) override;
+    CPLErr SetValue(int iRow, int iField, const void *pabyWKB,
+                    size_t nWKBSize) override;
 
     int ChangesAreWrittenToFile() override;
     void SetRowCount(int iCount) override;
